@@ -573,14 +573,61 @@ make_clone_setup (GnomeRRScreen *screen)
         return result;
 }
 
+static GnomeRRMode *
+find_best_mode (GnomeRROutput *output)
+{
+        GnomeRRMode *preferred;
+        GnomeRRMode **modes;
+        int best_size;
+        int best_width, best_height, best_rate;
+        int i;
+        GnomeRRMode *best_mode;
+
+        preferred = gnome_rr_output_get_preferred_mode (output);
+        if (preferred)
+                return preferred;
+
+        modes = gnome_rr_output_list_modes (output);
+        if (!modes)
+                return NULL;
+
+        best_size = best_width = best_height = best_rate = 0;
+        best_mode = NULL;
+
+        for (i = 0; modes[i] != NULL; i++) {
+                int w, h, r;
+                int size;
+
+                w = gnome_rr_mode_get_width (modes[i]);
+                h = gnome_rr_mode_get_height (modes[i]);
+                r = gnome_rr_mode_get_freq  (modes[i]);
+
+                size = w * h;
+
+                if (size > best_size) {
+                        best_size   = size;
+                        best_width  = w;
+                        best_height = h;
+                        best_rate   = r;
+                        best_mode   = modes[i];
+                } else if (size == best_size) {
+                        if (r > best_rate) {
+                                best_rate = r;
+                                best_mode = modes[i];
+                        }
+                }
+        }
+
+        return best_mode;
+}
+
 static gboolean
 turn_on (GnomeRRScreen *screen,
          GnomeOutputInfo *info,
          int x, int y)
 {
-        GnomeRROutput *output =
-                gnome_rr_screen_get_output_by_name (screen, info->name);
-        GnomeRRMode *mode = gnome_rr_output_get_preferred_mode (output);
+        GnomeRROutput *output = gnome_rr_screen_get_output_by_name (screen, info->name);
+        GnomeRRMode *mode = find_best_mode (output);
 
         if (mode) {
                 info->on = TRUE;
@@ -630,6 +677,21 @@ make_laptop_setup (GnomeRRScreen *screen)
 
 }
 
+static int
+turn_on_and_get_rightmost_offset (GnomeRRScreen *screen, GnomeOutputInfo *info, int x)
+{
+        if (info->on) {
+                info->x = x;
+                info->y = 0;
+                x += info->width;
+        } else {
+                if (turn_on (screen, info, x, 0))
+                        x += info->width;
+        }
+
+        return x;
+}
+
 static GnomeRRConfig *
 make_xinerama_setup (GnomeRRScreen *screen)
 {
@@ -644,21 +706,15 @@ make_xinerama_setup (GnomeRRScreen *screen)
         for (i = 0; result->outputs[i] != NULL; ++i) {
                 GnomeOutputInfo *info = result->outputs[i];
 
-                if (is_laptop (info)) {
-                        if (info->on || turn_on (screen, info, x, 0)) {
-                                x += info->width;
-                        }
-                }
+                if (is_laptop (info))
+                        x = turn_on_and_get_rightmost_offset (screen, info, x);
         }
 
         for (i = 0; result->outputs[i] != NULL; ++i) {
                 GnomeOutputInfo *info = result->outputs[i];
 
-                if (info->connected && !is_laptop (info)) {
-                        if (info->on || turn_on (screen, info, x, 0)) {
-                                x += info->width;
-                        }
-                }
+                if (info->connected && !is_laptop (info))
+                        x = turn_on_and_get_rightmost_offset (screen, info, x);
         }
 
         print_configuration (result, "xinerama setup");
@@ -695,7 +751,7 @@ make_other_setup (GnomeRRScreen *screen)
 }
 
 static GPtrArray *
-sanitize (GPtrArray *array)
+sanitize (GsdXrandrManager *manager, GPtrArray *array)
 {
         int i;
         GPtrArray *new;
@@ -712,10 +768,10 @@ sanitize (GPtrArray *array)
         /* Remove configurations that are duplicates of
          * configurations earlier in the cycle
          */
-        for (i = 1; i < array->len; ++i) {
+        for (i = 0; i < array->len; i++) {
                 int j;
 
-                for (j = 0; j < i; ++j) {
+                for (j = i + 1; j < array->len; j++) {
                         GnomeRRConfig *this = array->pdata[j];
                         GnomeRRConfig *other = array->pdata[i];
 
@@ -741,12 +797,34 @@ sanitize (GPtrArray *array)
                         }
 
                         if (all_off) {
+                                g_debug ("removing configuration as all outputs are off");
                                 gnome_rr_config_free (array->pdata[i]);
                                 array->pdata[i] = NULL;
                         }
                 }
         }
 
+        /* Do a final sanitization pass.  This will remove configurations that
+         * don't fit in the framebuffer's Virtual size.
+         */
+
+        for (i = 0; i < array->len; i++) {
+                GnomeRRConfig *config = array->pdata[i];
+
+                if (config) {
+                        GError *error;
+
+                        error = NULL;
+                        if (!gnome_rr_config_applicable (config, manager->priv->rw_screen, &error)) { /* NULL-GError */
+                                g_debug ("removing configuration which is not applicable because %s", error->message);
+                                g_error_free (error);
+
+                                gnome_rr_config_free (config);
+                                array->pdata[i] = NULL;
+                        }
+                }
+        }
+
         /* Remove NULL configurations */
         new = g_ptr_array_new ();
 
@@ -796,7 +874,7 @@ generate_fn_f7_configs (GsdXrandrManager *mgr)
         g_ptr_array_add (array, make_other_setup (screen));
         g_ptr_array_add (array, gnome_rr_config_new_stored (screen, NULL)); /* NULL-GError - if this can't read the stored config, no big deal */
 
-        array = sanitize (array);
+        array = sanitize (mgr, array);
 
         if (array) {
                 mgr->priv->fn_f7_configs = (GnomeRRConfig **)g_ptr_array_free (array, FALSE);
@@ -956,6 +1034,7 @@ auto_configure_outputs (GsdXrandrManager *manager, guint32 timestamp)
         GList *l;
         int x;
         GError *error;
+        gboolean applicable;
 
         config = gnome_rr_config_new_current (priv->rw_screen);
 
@@ -1028,12 +1107,47 @@ auto_configure_outputs (GsdXrandrManager *manager, guint32 timestamp)
                 x += output->width;
         }
 
+        /* Check if we have a large enough framebuffer size.  If not, turn off
+         * outputs from right to left until we reach a usable size.
+         */
+
+        just_turned_on = g_list_reverse (just_turned_on); /* now the outputs here are from right to left */
+
+        l = just_turned_on;
+        while (1) {
+                GnomeOutputInfo *output;
+                gboolean is_bounds_error;
+
+                error = NULL;
+                applicable = gnome_rr_config_applicable (config, priv->rw_screen, &error);
+
+                if (applicable)
+                        break;
+
+                is_bounds_error = g_error_matches (error, GNOME_RR_ERROR, GNOME_RR_ERROR_BOUNDS_ERROR);
+                g_error_free (error);
+
+                if (!is_bounds_error)
+                        break;
+
+                if (l) {
+                        i = GPOINTER_TO_INT (l->data);
+                        l = l->next;
+
+                        output = config->outputs[i];
+                        output->on = FALSE;
+                } else
+                        break;
+        }
+
         /* Apply the configuration! */
 
-        error = NULL;
-        if (!gnome_rr_config_apply_with_time (config, priv->rw_screen, timestamp, &error)) {
-                error_message (manager, _("Could not switch the monitor configuration"), error, NULL);
-                g_error_free (error);
+        if (applicable) {
+                error = NULL;
+                if (!gnome_rr_config_apply_with_time (config, priv->rw_screen, timestamp, &error)) {
+                        error_message (manager, _("Could not switch the monitor configuration"), error, NULL);
+                        g_error_free (error);
+                }
         }
 
         g_list_free (just_turned_on);
